//////////////////////////////////////////////////////////////
//
//  Whupsey - A Whups Client Library in C
//
//  $Horde: whupsey/test.c,v 1.6 2004/04/27 17:47:14 chuck Exp $
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "whupsey.h"

#define NELEMS(a) (sizeof(a) / sizeof(a[0]))

static char *statec[] = {"Unknown", "Unconfirmed", "New", "Assigned", "Resolved"};

void test_summary_ids     (Whupsey *w);
void test_create_ticket   (Whupsey *w);
void test_display_types   (Whupsey *w);
void test_display_queues (Whupsey *w);
void test_set_attributes  (Whupsey *w, unsigned int ticket_id, const WhupseyType *type);
void test_query           (Whupsey *w);

static char *server;
static char *user;
static char *password;
static int   create = 0;

void usage()
{
    printf("Usage: whupseytest [-c] server user password\n");
    printf("   eg. whupseytest http://example.com/horde/rpc.php jrandom password\n\n");
    printf("   '-c' will attempt to create a ticket in the remote whups server\n");
    printf("   otherwise, it will print the queues, types, states etc.\n\n");
    exit(1); 
}

void getargs(int argc, char **argv)
{
    int s = 0;

    if (argc > 1 && !strcmp(argv[1], "-c"))
        s = create = 1;

    if (argc < 3 + s)
    {
        server   = getenv("HORDE_SERVER");
        user     = getenv("HORDE_USER");
        password = getenv("HORDE_PASSWORD");

        if (!server || !user || !password)
            usage();
    }
    else
    {
        server   = argv[1 + s];
        user     = argv[2 + s];
        password = argv[3 + s];
    }
}

int main(int argc, char **argv)
{
    WhupseyResult  r;
    Whupsey       *w;

    getargs(argc, argv);

    printf("Attempting to connect to whups server:\n");
    printf("   Server: %s\n", server);
    printf("   User:   %s\n", user);
    printf("\n");

    whupsey_create(&w, "WhupseyTest", "1.0", server, user, password);

    r = whupsey_load_schema(w);
    if (r != WHUPSEY_OK)
    {
        if (r == WHUPSEY_RPCFAULT)
            printf("Error: Couldn't load schema: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
        else if (r == WHUPSEY_NOMEMORY)
            printf("Error: Couldn't load schema: Out of memory\n");
        else
            printf("Error: Couldn't load schema: Unknown error\n");

        exit(1);
    }

    test_display_types(w);
    test_display_queues(w);
//    test_query(w);

    if (create)
        test_create_ticket(w);

    whupsey_free(w);
    return 0;
}

void display_ids(Whupsey *w, const char *title, int *ids, int c)
{
    int i;

    printf("%s:\n", title);

    if (c == 0)
    {
        printf("    None\n");
    }
    else
    {
        for (i = 0; i < c; i++)
            printf("    %d\n", ids[i]);
    }

    printf("\n");
}

void test_summary_ids(Whupsey *w)
{
    int *ids;
    int  c;

    if (whupsey_get_assigned_tickets(w, &ids, &c) == WHUPSEY_RPCFAULT)
    {
        printf("Error: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
    }
    else
    {
        display_ids(w, "Assigned Tickets", ids, c);
        free(ids);
    }

    if (whupsey_get_requested_tickets(w, &ids, &c) == WHUPSEY_RPCFAULT)
    {
        printf("Error: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
    }
    else
    {
        display_ids(w, "Requested Tickets", ids, c);
        free(ids);
    }
}

void test_create_ticket(Whupsey *w)
{
    WhupseyAttribute attribs[2];
    WhupseyTicket    tickets[2];

    const WhupseyQueue  *queues = NULL;
    const WhupseyVersion *version = NULL;
    const WhupseyState   *state   = NULL;
    const WhupseyType    *types   = NULL;

    int nqueues;
    int ntypes;
    int count;
    int i;

    printf("-----------------------------\n");
    printf("Attempting to create tickets:\n");

    whupsey_get_queues (w, &queues, &nqueues);
    whupsey_get_types   (w, &types,   &ntypes);

    if (nqueues == 0)
    {
        printf("  Can't create ticket - no queues available\n");
        return;
    }

    if (queues->ntypes == 0)
    {
        printf("  Can't create ticket - no types associated with the first queue\n");
        return;
    }

    if (queues->types[0]->npriorities == 0)
    {
        printf("  Can't create ticket - no priorities associated with type %s\n", queues->types[0]->name);
        return;
    }

    state = queues[0].types[0]->states;
    count = queues[0].types[0]->nstates;
    for (i = 0; i < count; i++, state++)
        if (state->category == WHUPSEY_UNCONFIRMED)
            break;

    if (i == count)
    {
        state = queues[0].types[0]->states;
        count = queues[0].types[0]->nstates;
        for (i = 0; i < count; i++, state++)
            if (state->category == WHUPSEY_NEW)
                break;

        if (i == count)
        {
            printf("  Can't create ticket - no unconfirmed or new states in type %s\n", queues[0].types[0]->name);
            return;
        }
    }

    if (queues[0].has_versions)
    {
        if (queues[0].nversions == 0)
        {
            printf("  Can't create ticket - queue %s is versioned but with no versions defined\n", queues[0].name);
            return;
        }

        version = queues[0].versions;
    }

    printf("  Queue:    %s\n", queues[0].name);
    printf("  Type:      %s\n", queues[0].types[0]->name);
    printf("  State:     %s\n", state->name);
    printf("  Priority:  %s\n", queues[0].types[0]->priorities[0].name);
    printf("  Version:   %s\n", version ? version->name : "(No versions)");

    attribs[0].descriptor = queues[0].types[0]->attributes;

    if (attribs[0].descriptor != NULL)
    {
        attribs[0].value       = "Test Attribute";
        tickets[0].attributes  = attribs;
        tickets[0].nattributes = 1;

        printf("  Attribute: %s => %s\n", attribs[0].descriptor->name, attribs[0].value);
    }

    printf("\n");

    tickets[0].id         = 0; // will be set by whupsey_add_ticket (if successful)
    tickets[0].comment_id = 0; //   "
    tickets[0].timestamp  = 0; //   "

    tickets[0].type     = queues[0].types[0];
    tickets[0].state    = state;
    tickets[0].priority = &queues[0].types[0]->priorities[0];
    tickets[0].queue   = &queues[0];
    tickets[0].version  = NULL;

    tickets[0].requester = NULL;
    tickets[0].owner     = NULL;
    tickets[0].summary   = "RPC Generated Ticket";
    tickets[0].comment   = "This ticket was created by the WhupseyTest whups client.\n\nTesty test.\n";

    tickets[1] = tickets[0];

    if (whupsey_add_tickets(w, tickets, NELEMS(tickets)) != WHUPSEY_OK)
        printf("Error: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
    else
    {
        for (i = 0; i < NELEMS(tickets); i++)
        {
            printf("New ticket id: %d, New comment id: %d, Timestamp %ld\n", tickets[i].id, tickets[i].comment_id, tickets[i].timestamp);

//            printf("\nTrying to set the attributes:\n");
//            test_set_attributes(w, tickets[i].id, tickets[i].type);
        }
    }

    printf("\n");
}

void test_display_types(Whupsey *w)
{
    int ntypes;
    const WhupseyType *types;

    int i, j;

    whupsey_get_types(w, &types, &ntypes);

    printf("Got %d types:\n", ntypes);
    for (i = 0; i < ntypes; i++)
    {
        printf("    Id:         %d\n", types[i].id);
        printf("    Name:       %s\n", types[i].name);
        printf("    Desc:       %s\n", types[i].description);
        printf("    States:     %d states for this type:", types[i].nstates);

        for (j = 0; j < types[i].nstates; j++)
        {
            printf("\n");
            printf("        Id:       %d\n", types[i].states[j].id);
            printf("        Name:     %s\n", types[i].states[j].name);
            printf("        Desc:     %s\n", types[i].states[j].description);
            printf("        Category: %s\n", statec[types[i].states[j].category]);
        }

        printf("\n    Priorities: %d priorities for this type:\n", types[i].npriorities);

        for (j = 0; j < types[i].npriorities; j++)
        {
            printf("        Id:       %d\n", types[i].priorities[j].id);
            printf("        Name:     %s\n", types[i].priorities[j].name);
            printf("        Desc:     %s\n", types[i].priorities[j].description);
            printf("\n");
        }

        if (!types[i].npriorities)
            printf("\n");

        printf("    Attributes: %d attributes for this type:\n", types[i].nattributes);

        for (j = 0; j < types[i].nattributes; j++)
        {
            printf("        Id:       %d\n", types[i].attributes[j].id);
            printf("        Name:     %s\n", types[i].attributes[j].name);
            printf("        Desc:     %s\n", types[i].attributes[j].description);
            printf("\n");
        }

        if (!types[i].nattributes)
            printf("\n");
    }

    printf("\n");
}

void test_display_queues(Whupsey *w)
{
    int nqueues;
    const WhupseyQueue *queues;

    int i, j;

    whupsey_get_queues(w, &queues, &nqueues);

    printf("Got %d queues:\n", nqueues);
    for (i = 0; i < nqueues; i++)
    {
        printf("    Id:        %d\n", queues[i].id);
        printf("    Name:      %s\n", queues[i].name);
        printf("    Desc:      %s\n", queues[i].description);
        printf("    Versioned: %s\n", queues[i].has_versions ? "yes" : "no");
        printf("    Subjects:  %s\n", queues[i].has_subjects ? "yes" : "no");
        printf("    Types:     %d types for this queue:\n", queues[i].ntypes);

        for (j = 0; j < queues[i].ntypes; j++)
            printf("        Type %d (%s)\n", queues[i].types[j]->id, queues[i].types[j]->name);

        if (queues[i].has_versions)
        {
            printf("    Versions:  %d versions for this queue:\n", queues[i].nversions);

            for (j = 0; j < queues[i].ntypes; j++)
            {
                printf("        Id:   %d\n", queues[i].versions[j].id);
                printf("        Name: %s\n", queues[i].versions[j].name);
                printf("        Desc: %s\n", queues[i].versions[j].description);
                printf("\n");
            }
        }

        printf("\n");
    }
}

void test_set_attributes(Whupsey *w, unsigned int ticket_id, const WhupseyType *type)
{
    // A bit hardcoded at the moment, so not run by default

    WhupseyAttribute attribs[2];

    attribs[0].descriptor = whupsey_find_attribute_desc_by_name(w, type, "Test Attribute 1");
    attribs[1].descriptor = whupsey_find_attribute_desc_by_name(w, type, "Test Attribute 2");

    if (!attribs[0].descriptor || !attribs[1].descriptor)
    {
        printf("test_set_attributes: can't find attribute descriptors\n");
        return;
    }

    attribs[0].value = "Test 1";
    attribs[1].value = "Wibbly wibbly woo";

    if (whupsey_set_ticket_attributes(w, ticket_id, "Mr. Comment\n", attribs, 2) != WHUPSEY_OK)
        printf("Error setting attributes: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
    else
        printf("Successfully set the ticket attributes\n");
}

void test_query(Whupsey *w)
{
    // A bit hardcoded at the moment, so not run by default

    WhupseyQueryParameter params[2] = {{"Param1", "Param1 Value"}, {"Param2", "Param2 Value"}};
    WhupseyQuery queries[2];

    memset(&queries, 0, sizeof(queries));

    queries[0].query_name  = "My Requested Tickets";
    queries[0].get_details = 0;
    queries[0].parameters  = params;
    queries[0].nparameters = 2;

    queries[1].query_name  = "My Tickets";
    queries[1].get_details = 1;
    queries[1].parameters  = NULL;
    queries[1].nparameters = 0;

    if (whupsey_run_queries(w, queries, NELEMS(queries)) != WHUPSEY_OK)
    {
        printf("Error querying tickets: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
    }
    else
    {
        int i, j;

        printf("Successfully queried for some tickets:\n");

        for (i = 0; i < NELEMS(queries); i++)
        {
            printf("    %d results for %s:\n", queries[i].ntickets, queries[i].query_name);

            if (queries[i].get_details)
            {
                WhupseyTicket *t = queries[i].tickets;

                for (j = 0; j < queries[i].ntickets; j++)
                {
                    printf("        Id:        %d\n", t[j].id);
                    printf("        Type:      %s\n", t[j].type->name);
                    printf("        State:     %s\n", t[j].state->name);
                    printf("        Priority:  %s\n", t[j].priority->name);
                    printf("        Queue:     %s\n", t[j].queue->name);
                    printf("        Version:   %s\n", t[j].version ? t[j].version->name : "(none)");
                    printf("        Summary:   %s\n", t[j].summary);
                    printf("        Requester: %s\n", t[j].requester);
                    printf("        Owner:     %s\n", t[j].owner);
                    printf("\n");
                }
            }
            else
            {
                for (j = 0; j < queries[i].ntickets; j++)
                    printf("        %d\n", queries[i].ticket_ids[j]);

                printf("\n");
            }
        }

        whupsey_clear_queries(w, queries, NELEMS(queries));
    }
}

