/////////////////////////////////////////////////////////////////
//
//  Whupsey - A Whups Client Library in C
//
//  $Horde: whupsey/whupsey.c,v 1.5 2004/04/27 17:45:52 chuck Exp $
//

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "whupsey.h"

static void whupsey_free_schema     (Whupsey *w);
static void whupsey_free_types      (Whupsey *w);
static void whupsey_free_states     (Whupsey *w, WhupseyState *states, int nstates);
static void whupsey_free_priorities (Whupsey *w, WhupseyPriority *priorities, int npriorities);
static void whupsey_free_queues     (Whupsey *w);

struct Whupsey
{
    xmlrpc_env          env;
    xmlrpc_server_info *server;

    int ntypes;
    WhupseyType *types;

    int nqueues;
    WhupseyQueue *queues;
};

static inline void CleanWhupsey(Whupsey *w)
{
    if (w->env.fault_occurred)
    {
        xmlrpc_env_clean(&w->env);

        w->env.fault_occurred = 0;
        w->env.fault_code     = 0;
        w->env.fault_string   = NULL;
    }
}

static inline void DumpValue(Whupsey *w, xmlrpc_value *v)
{
    int s;
    xmlrpc_mem_block *b;
    CleanWhupsey(w);
    b = xmlrpc_mem_block_new(&w->env, 0);
    xmlrpc_serialize_value(&w->env, b, v);
    s = xmlrpc_mem_block_size(b);
    printf("Result:\n%*.*s\n\n", s, s, (char *) xmlrpc_mem_block_contents(b));
    xmlrpc_mem_block_free(b);
}


// Simple accessor methods first

int whupsey_get_fault_code(Whupsey *w)
{
    return w->env.fault_code;
}

const char *whupsey_get_fault_string(Whupsey *w)
{
    return w->env.fault_string;
}

void whupsey_get_types(Whupsey *w, const WhupseyType **types, int *ntypes)
{
    *ntypes = w->ntypes;
    *types  = w->types;
}

void whupsey_get_queues(Whupsey *w, const WhupseyQueue **queues, int *nqueues)
{
    *nqueues = w->nqueues;
    *queues  = w->queues;
}

const WhupseyType *whupsey_find_type_by_id(Whupsey *w, unsigned int id)
{
    int i;

    for (i = 0; i < w->ntypes; i++)
        if (w->types[i].id == id)
            return &w->types[i];

    return NULL;
}

const WhupseyType *whupsey_find_type_by_name(Whupsey *w, const char *name)
{
    int i;

    for (i = 0; i < w->ntypes; i++)
        if (!strcmp(w->types[i].name, name))
            return &w->types[i];

    return NULL;
}

const WhupseyState *whupsey_find_state_by_id(Whupsey *w, const WhupseyType *type, unsigned int state_id)
{
    int i;

    for (i = 0; i < type->nstates; i++)
        if (type->states[i].id == state_id)
            return &type->states[i];

    return NULL;
}

const WhupseyQueue *whupsey_find_queue_by_id(Whupsey *w, unsigned int queue_id)
{
    int i;

    for (i = 0; i < w->nqueues; i++)
        if (w->queues[i].id == queue_id)
            return &w->queues[i];

    return NULL;
}

const WhupseyPriority *whupsey_find_priority_by_id(Whupsey *w, const WhupseyType *type, unsigned int priority_id)
{
    int i;

    for (i = 0; i < type->npriorities; i++)
        if (type->priorities[i].id == priority_id)
            return &type->priorities[i];

    return NULL;
}

const WhupseyAttributeDesc *whupsey_find_attribute_desc_by_id(Whupsey *w, const WhupseyType *type, unsigned int id)
{
    int i;

    for (i = 0; i < type->nattributes; i++)
        if (type->attributes[i].id == id)
            return &type->attributes[i];

    return NULL;
}

const WhupseyAttributeDesc *whupsey_find_attribute_desc_by_name(Whupsey *w, const WhupseyType *type, const char *name)
{
    int i;

    for (i = 0; i < type->nattributes; i++)
        if (!strcmp(type->attributes[i].name, name))
            return &type->attributes[i];

    return NULL;
}


// Now for the real stuff

WhupseyResult whupsey_create(Whupsey **result, const char *useragent, const char *version, const char *server, const char *user, const char *pass)
{
    Whupsey *w = calloc(1, sizeof(Whupsey));

    *result = NULL;

    if (w == NULL)
        return WHUPSEY_NOMEMORY;

    xmlrpc_client_init(XMLRPC_CLIENT_NO_FLAGS, (char *) useragent, (char *) version);
    xmlrpc_env_init(&w->env);

    w->server = xmlrpc_server_info_new(&w->env, (char *) server);
    if (!w->env.fault_occurred)
        xmlrpc_server_info_set_basic_auth(&w->env, w->server, (char *) user, (char *) pass);

    if (w->env.fault_occurred)
    {
        whupsey_free(w);
        return WHUPSEY_RPCFAULT;
    }

    *result = w;
    return WHUPSEY_OK;
}

static WhupseyResult whupsey_get_int_array(Whupsey *w, const char *method,  int **result, int *count)
{
    xmlrpc_value *ids;
    xmlrpc_value *id;
    int size, i;

    CleanWhupsey(w);

    *result = NULL;
    *count  = 0;

    ids = xmlrpc_client_call_server(&w->env, w->server, (char *) method, "()");
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    size = xmlrpc_array_size(&w->env, ids);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    *result = malloc(sizeof(int) * size);
    if (!*result)
        return WHUPSEY_NOMEMORY;

    for (i = 0; i < size; i++)
    {
        id = xmlrpc_array_get_item(&w->env, ids, i);
        if (!w->env.fault_occurred)
            xmlrpc_parse_value(&w->env, id, "i", *result + i);

        if (w->env.fault_occurred)
        {
            xmlrpc_DECREF(ids);
            free(*result);
            *result = NULL;
            return WHUPSEY_RPCFAULT;
        }
    }

    xmlrpc_DECREF(ids);
    *count = size;
    return WHUPSEY_OK;
}

WhupseyResult whupsey_get_assigned_tickets(Whupsey *w, int **result, int *count)
{
    return whupsey_get_int_array(w, "projects.get_assigned_ticket_ids", result, count);
}

WhupseyResult whupsey_get_requested_tickets(Whupsey *w, int **result, int *count)
{
    return whupsey_get_int_array(w, "projects.get_requested_ticket_ids", result, count);
}


///////////////////// Add Tickets

static WhupseyResult build_add_ticket_attributes(Whupsey *w, WhupseyTicket *ticket, xmlrpc_value **result)
{
    xmlrpc_value *value = NULL;
    xmlrpc_value *entry = NULL;
    int i;

    *result = NULL;

    value = xmlrpc_build_value(&w->env, "()");
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    if (ticket->nattributes < 1)
    {
        *result = value;
        return WHUPSEY_OK;
    }

    for (i = 0; i < ticket->nattributes; i++)
    {
        entry = xmlrpc_build_value(
            &w->env,
            "{s:i,s:s}",
            "id",    ticket->attributes[i].descriptor->id,
            "value", ticket->attributes[i].value);
        if (w->env.fault_occurred)
            goto fail;

        xmlrpc_array_append_item(&w->env, value, entry);
        xmlrpc_DECREF(entry);

        if (w->env.fault_occurred)
            goto fail;
    }

    *result = value;
    return WHUPSEY_OK;

fail:
    if (value != NULL)
        xmlrpc_DECREF(value);

    return WHUPSEY_RPCFAULT;
}

static WhupseyResult build_add_ticket_struct(Whupsey *w, WhupseyTicket *ticket, xmlrpc_value **result)
{
    WhupseyResult r;
    xmlrpc_value *attributes;
    xmlrpc_value *value;

    *result = NULL;

    if (ticket->type == NULL || ticket->state == NULL || ticket->priority == NULL || ticket->queue == NULL)
        return WHUPSEY_BADARG;

    r = build_add_ticket_attributes(w, ticket, &attributes);
    if (r != WHUPSEY_OK)
        return r;

    value = xmlrpc_build_value(
        &w->env,
        "{s:i,s:i,s:i,s:i,s:i,s:s,s:s,s:s,s:V}",
        "type",       ticket->type->id,
        "state",      ticket->state->id,
        "priority",   ticket->priority->id,
        "queue",     ticket->queue->id,
        "version",    ticket->version ? ticket->version->id : 0,
        "owner",      ticket->owner ? ticket->owner : "",
        "summary",    ticket->summary,
        "comment",    ticket->comment,
        "attributes", attributes);

    xmlrpc_DECREF(attributes);

    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    *result = value;
    return WHUPSEY_OK;
}

static WhupseyResult parse_add_ticket_result(Whupsey *w, xmlrpc_value *value, WhupseyTicket *tickets, int ntickets)
{
    xmlrpc_value *entry;
    int size;
    int i;

    int  id;
    int  comment_id;
    long timestamp;

    WhupseyResult r = WHUPSEY_OK;

    size = xmlrpc_array_size(&w->env, value);
    if (w->env.fault_occurred || size != ntickets)
        return WHUPSEY_RPCFAULT;

    for (i = 0; i < ntickets; i++)
    {
        entry = xmlrpc_array_get_item(&w->env, value, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, entry,
            "{s:i,s:i,s:i,*}",
            "id",         &id,
            "comment_id", &comment_id,
            "timestamp",  &timestamp);

        if (w->env.fault_occurred)
        {
            r = WHUPSEY_RPCFAULT;
        }
        else
        {
            tickets[i].id         = id;
            tickets[i].comment_id = comment_id;
            tickets[i].timestamp  = timestamp;
        }
    }

    return r;
}

WhupseyResult whupsey_add_tickets(Whupsey *w, WhupseyTicket *tickets, int ntickets)
{
    xmlrpc_value *ticket_array = NULL;
    xmlrpc_value *entry        = NULL;
    xmlrpc_value *result       = NULL;

    WhupseyResult r = WHUPSEY_OK;
    int i;

    if (!w || !tickets || !ntickets)
        return WHUPSEY_BADARG;

    CleanWhupsey(w);

    ticket_array = xmlrpc_build_value(&w->env, "()");
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    for (i = 0; i < ntickets && r == WHUPSEY_OK; i++)
    {
        r = build_add_ticket_struct(w, &tickets[i], &entry);
        if (r == WHUPSEY_OK)
        {
            xmlrpc_array_append_item(&w->env, ticket_array, entry);
            if (w->env.fault_occurred)
                r = WHUPSEY_RPCFAULT;

            xmlrpc_DECREF(entry);
            entry = NULL;
        }
    }

    if (r == WHUPSEY_OK)
    {
        result = xmlrpc_client_call_server(&w->env, w->server, "projects.add_tickets", "(V)", ticket_array);
        if (w->env.fault_occurred)
            r = WHUPSEY_RPCFAULT;
        else
            r = parse_add_ticket_result(w, result, tickets, ntickets);
    }

    if (result)       xmlrpc_DECREF(result);
    if (ticket_array) xmlrpc_DECREF(ticket_array);
    if (entry)        xmlrpc_DECREF(entry);

    return r;
}

WhupseyResult whupsey_set_ticket_attributes(Whupsey *w, unsigned int ticket_id, const char *comment, WhupseyAttribute *attribs, int nattribs)
{
    xmlrpc_value *attrib_array;
    xmlrpc_value *entry;
    xmlrpc_value *result;
    int comment_id;
    int i;

    CleanWhupsey(w);

    if (nattribs == 0)
        return WHUPSEY_OK;

    if (!w || !ticket_id || attribs == NULL)
        return WHUPSEY_BADARG;

    attrib_array = xmlrpc_build_value(&w->env, "()");
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    for (i = 0; i < nattribs; i++)
    {
        entry = xmlrpc_build_value(&w->env,
                    "{s:i,s:s}",
                    "attribute_id",    attribs[i].descriptor->id,
                    "attribute_value", attribs[i].value);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_array_append_item(&w->env, attrib_array, entry);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;
    }

    result = xmlrpc_client_call_server(
        &w->env, w->server, "projects.set_ticket_attributes",
        "({s:i,s:s,s:V})",
        "ticket_id",  ticket_id,
        "comment",    comment,
        "attributes", attrib_array);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    xmlrpc_parse_value(&w->env, result, "i", &comment_id);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    return WHUPSEY_OK;
}


///////////////////// Queries

static WhupseyResult build_query_struct(Whupsey *w, WhupseyQuery *query, xmlrpc_value **result)
{
    xmlrpc_value *param_struct;
    xmlrpc_value *value;
    xmlrpc_value *newvalue;
    int i;

    *result = NULL;

    param_struct = xmlrpc_struct_new(&w->env);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    for (i = 0; i < query->nparameters; i++)
    {
        value = xmlrpc_build_value(&w->env, "s", query->parameters[i].value);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_struct_set_value(&w->env, param_struct, query->parameters[i].name, value);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;
    }

    newvalue = xmlrpc_build_value(&w->env,
        "{s:s,s:b,s:V}",
        "query_name",       query->query_name,
        "query_details",    (xmlrpc_bool) query->get_details,
        "query_parameters", param_struct);
    if (w->env.fault_occurred)
    {
        xmlrpc_DECREF(param_struct);
        return WHUPSEY_RPCFAULT;
    }

    *result = newvalue;
    return WHUPSEY_OK;
}

static WhupseyResult parse_int_array(Whupsey *w, xmlrpc_value *intarray, int **out, int *nout)
{
    xmlrpc_value *element;
    int size;
    int i;

    *out  = NULL;
    *nout = 0;

    size = xmlrpc_array_size(&w->env, intarray);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    *out = malloc(sizeof(int) * size);
    if (!*out)
        return WHUPSEY_NOMEMORY;

    for (i = 0; i < size; i++)
    {
        element = xmlrpc_array_get_item(&w->env, intarray, i);
        if (!w->env.fault_occurred)
            xmlrpc_parse_value(&w->env, element, "i", *out + i);

        if (w->env.fault_occurred)
        {
            free(*out);
            *out = NULL;
            return WHUPSEY_RPCFAULT;
        }
    }

    *nout = size;
    return WHUPSEY_OK;
}

static WhupseyResult parse_ticket(Whupsey *w, xmlrpc_value *tvalue, WhupseyTicket *ticket)
{
    char *summary;
    char *requester;
    char *owner;
    int   type_id;
    int   state_id;
    int   queue_id;
    int   priority_id;

    xmlrpc_parse_value(
        &w->env, tvalue,
        "{s:i,s:s,s:s,s:s,s:i,s:i,s:i,s:i,*}",
        "id",                &ticket->id,
        "summary",           &summary,
        "user_id_requester", &requester,
        "user_id_owner",     &owner,
        "state",             &state_id,
        "type",              &type_id,
        "priority",          &priority_id,
        "queue",            &queue_id,
        "comment_id",        &ticket->comment_id);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    ticket->type = whupsey_find_type_by_id(w, type_id);
    if (ticket->type == NULL)
        return WHUPSEY_RPCFAULT;  // FIXME: Bad error code (doesn't describe the actual error)

    ticket->state    = whupsey_find_state_by_id(w, ticket->type, state_id);
    ticket->queue   = whupsey_find_queue_by_id(w, queue_id);
    ticket->priority = whupsey_find_priority_by_id(w, ticket->type, priority_id);

    if (ticket->state == NULL || ticket->queue == NULL || ticket->priority == NULL)
        return WHUPSEY_RPCFAULT;  // FIXME: Bad error code

    ticket->summary   = strdup(summary);
    ticket->requester = strdup(requester);
    ticket->owner     = strdup(owner);

    return WHUPSEY_OK;
}

static WhupseyResult parse_ticket_array(Whupsey *w, xmlrpc_value *tarray, WhupseyTicket **tickets, int *ntickets)
{
    WhupseyResult r;
    xmlrpc_value *element;
    int size;
    int i;

    *tickets  = NULL;
    *ntickets = 0;

    size = xmlrpc_array_size(&w->env, tarray);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    *tickets = calloc(size, sizeof(WhupseyTicket));
    if (!*tickets)
        return WHUPSEY_NOMEMORY;

    for (i = 0; i < size; i++)
    {
        element = xmlrpc_array_get_item(&w->env, tarray, i);
        if (!w->env.fault_occurred)
            r = parse_ticket(w, element, *tickets + i);

        if (w->env.fault_occurred)
        {
            free(*tickets);
            *tickets = NULL;
            return WHUPSEY_RPCFAULT;
        }
    }

    *ntickets = size;
    return WHUPSEY_OK;
}

WhupseyResult whupsey_run_queries(Whupsey *w, WhupseyQuery *queries, int nqueries)
{
    xmlrpc_value *query_info;
    xmlrpc_value *query_array;
    xmlrpc_value *result;
    xmlrpc_value *qresult;
    WhupseyResult r;
    int size;
    int i, j;

    if (!w || !queries || !nqueries)
        return WHUPSEY_BADARG;

    CleanWhupsey(w);

    // Build arguments
    query_array = xmlrpc_build_value(&w->env, "()");
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    for (i = 0; i < nqueries; i++)
    {
        r = build_query_struct(w, &queries[i], &query_info);

        if (r == WHUPSEY_OK)
        {
            xmlrpc_array_append_item(&w->env, query_array, query_info);
            if (w->env.fault_occurred)
                r = WHUPSEY_RPCFAULT;
        }

        if (r != WHUPSEY_OK)
        {
            xmlrpc_DECREF(query_array);
            return r;
        }
    }

    // Call procedure
    result = xmlrpc_client_call_server(&w->env, w->server, "projects.query", "(V)", query_array);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    r = WHUPSEY_OK;

    // Parse results
    size = xmlrpc_array_size(&w->env, result);
    if (w->env.fault_occurred || size != nqueries)
        r = WHUPSEY_RPCFAULT;
    else
    {
        for (i = 0; i < nqueries && r == WHUPSEY_OK; i++)
        {
            qresult = xmlrpc_array_get_item(&w->env, result, i);
            if (w->env.fault_occurred)
                r = WHUPSEY_RPCFAULT;
            else
            {
                if (queries[i].get_details)
                    r = parse_ticket_array(w, qresult, &queries[i].tickets, &queries[i].ntickets);
                else
                    r = parse_int_array(w, qresult, &queries[i].ticket_ids, &queries[i].ntickets);
            }
        }
    }

    // Clean up after any errors
    if (r != WHUPSEY_OK)
    {
        for (j = 0; j < i; j++)
        {
            if (queries[j].get_details)
                free(queries[j].tickets);
            else
                free(queries[j].ticket_ids);

            queries[j].ntickets = 0;
        }
    }

    xmlrpc_DECREF(result);
    return r;
}


///////////////////// Schema loading

static WhupseyResult whupsey_parse_states(Whupsey *w, xmlrpc_value *states, WhupseyType *type)
{
    char *name;
    char *description;
    char *category;
    int size, i;

    size = xmlrpc_array_size(&w->env, states);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    type->states  = calloc(size, sizeof(WhupseyState));
    type->nstates = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *state = xmlrpc_array_get_item(&w->env, states, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, state,
            "{s:i,s:s,s:s,s:s,*}",
            "state_id",          &type->states[i].id,
            "state_name",        &name,
            "state_description", &description,
            "state_category",    &category);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        type->states[i].name        = strdup(name);
        type->states[i].description = strdup(description);

        if (!strcasecmp(category, "unconfirmed"))
            type->states[i].category = WHUPSEY_UNCONFIRMED;
        else if (!strcasecmp(category, "new"))
            type->states[i].category = WHUPSEY_NEW;
        else if (!strcasecmp(category, "assigned"))
            type->states[i].category = WHUPSEY_ASSIGNED;
        else if (!strcasecmp(category, "resolved"))
            type->states[i].category = WHUPSEY_RESOLVED;
        else
            // Return this error to the user somehow (set w->env.fault_code?)
            return WHUPSEY_RPCFAULT;
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_priorities(Whupsey *w, xmlrpc_value *priorities, WhupseyType *type)
{
    char *name;
    char *description;
    int size, i;

    size = xmlrpc_array_size(&w->env, priorities);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    type->priorities  = calloc(size, sizeof(WhupseyPriority));
    type->npriorities = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *priority = xmlrpc_array_get_item(&w->env, priorities, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, priority,
            "{s:i,s:s,s:s,*}",
            "priority_id",          &type->priorities[i].id,
            "priority_name",        &name,
            "priority_description", &description);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        type->priorities[i].name        = strdup(name);
        type->priorities[i].description = strdup(description);
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_attribute_descs(Whupsey *w, xmlrpc_value *attribs, WhupseyType *type)
{
    char *name;
    char *description;
    int size, i;

    size = xmlrpc_array_size(&w->env, attribs);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    type->attributes  = calloc(size, sizeof(WhupseyAttributeDesc));
    type->nattributes = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *attrib = xmlrpc_array_get_item(&w->env, attribs, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, attrib,
            "{s:i,s:s,s:s,*}",
            "attribute_id",          &type->attributes[i].id,
            "attribute_name",        &name,
            "attribute_description", &description);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        type->attributes[i].name        = strdup(name);
        type->attributes[i].description = strdup(description);
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_types(Whupsey *w, xmlrpc_value *types)
{
    WhupseyResult r;
    xmlrpc_value *states;
    xmlrpc_value *priorities;
    xmlrpc_value *attributes;
    char *name;
    char *description;
    int size, i;

    size = xmlrpc_array_size(&w->env, types);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    w->types = calloc(size, sizeof(WhupseyType));
    if (!w->types)
        return WHUPSEY_NOMEMORY;

    w->ntypes = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *type = xmlrpc_array_get_item(&w->env, types, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, type,
            "{s:i,s:s,s:s,s:A,s:A,s:A,*}",
            "type_id",          &w->types[i].id,
            "type_name",        &name,
            "type_description", &description,
            "type_states",      &states,
            "type_priorities",  &priorities,
            "type_attributes",  &attributes);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        w->types[i].name        = strdup(name);
        w->types[i].description = strdup(description);

        r = whupsey_parse_states(w, states, w->types + i);
        if (r != WHUPSEY_OK)
            return r;

        r = whupsey_parse_priorities(w, priorities, w->types + i);
        if (r != WHUPSEY_OK)
            return r;

        r = whupsey_parse_attribute_descs(w, attributes, w->types + i);
        if (r != WHUPSEY_OK)
            return r;
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_type_ids(Whupsey *w, xmlrpc_value *type_ids, WhupseyQueue *m)
{
    unsigned int id;
    int size, i;

    size = xmlrpc_array_size(&w->env, type_ids);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    m->types = calloc(size, sizeof(WhupseyType *));
    if (!m->types)
        return WHUPSEY_NOMEMORY;

    m->ntypes = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *type = xmlrpc_array_get_item(&w->env, type_ids, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(&w->env, type, "i", &id);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        m->types[i] = (WhupseyType *) whupsey_find_type_by_id(w, id);
        if (m->types[i] == NULL)
            return WHUPSEY_RPCFAULT; // FIXME: Better way to return this error
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_versions(Whupsey *w, xmlrpc_value *versions, WhupseyQueue *m)
{
    char *name;
    char *description;
    int size, i;

    size = xmlrpc_array_size(&w->env, versions);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    if (size == 0)
        return WHUPSEY_OK;

    m->versions = calloc(size, sizeof(WhupseyVersion));
    if (!m->versions)
        return WHUPSEY_NOMEMORY;

    m->nversions = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *version = xmlrpc_array_get_item(&w->env, versions, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, version,
            "{s:i,s:s,s:s,*}",
            "version_id",          &m->versions[i].id,
            "version_name",        &name,
            "version_description", &description);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        m->versions[i].name        = strdup(name);
        m->versions[i].description = strdup(description);
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_queues(Whupsey *w, xmlrpc_value *queues)
{
    WhupseyResult r;
    xmlrpc_value *versions;
    xmlrpc_value *type_ids;

    char *name;
    char *description;

    int size, i;

    size = xmlrpc_array_size(&w->env, queues);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    w->queues = calloc(size, sizeof(WhupseyQueue));
    if (!w->queues)
        return WHUPSEY_NOMEMORY;

    w->nqueues = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *queue = xmlrpc_array_get_item(&w->env, queues, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, queue,
            "{s:i,s:s,s:s,s:i,s:i,s:A,s:A,*}",
            "queue_id",          &w->queues[i].id,
            "queue_name",        &name,
            "queue_description", &description,
            "queue_subjectlist", &w->queues[i].has_subjects,
            "queue_versioned",   &w->queues[i].has_versions,
            "queue_versions",    &versions,
            "queue_type_ids",    &type_ids);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        w->queues[i].name        = strdup(name);
        w->queues[i].description = strdup(description);

        r = whupsey_parse_type_ids(w, type_ids, w->queues + i);
        if (r != WHUPSEY_OK)
            return r;

        r = whupsey_parse_versions(w, versions, w->queues + i);
        if (r != WHUPSEY_OK)
            return r;
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_schema(Whupsey *w, xmlrpc_value *schema)
{
    WhupseyResult  r;
    xmlrpc_value  *types;
    xmlrpc_value  *queues;

    xmlrpc_parse_value(
        &w->env, schema,
        "{s:A,s:A,*}",
        "types",   &types,
        "queues", &queues);

    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    r = whupsey_parse_types(w, types);
    if (r != WHUPSEY_OK)
        return r;

    r = whupsey_parse_queues(w, queues);
    if (r != WHUPSEY_OK)
        return r;

    return WHUPSEY_OK;
}

WhupseyResult whupsey_load_schema(Whupsey *w)
{
    xmlrpc_value  *schema = NULL;
    WhupseyResult  r;

    CleanWhupsey(w);

    if (!w)
        return WHUPSEY_BADARG;

    whupsey_free_schema(w);

    schema = xmlrpc_client_call_server(&w->env, w->server, "projects.get_schema", "()");
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    r = whupsey_parse_schema(w, schema);
    if (r != WHUPSEY_OK)
        whupsey_free_schema(w);

    xmlrpc_DECREF(schema);
    return r;
}

void whupsey_free(Whupsey *w)
{
    whupsey_free_schema(w);
    xmlrpc_env_clean(&w->env);

    if (w->server)
        xmlrpc_server_info_free(w->server);

    xmlrpc_client_cleanup();
    free(w);
}

void whupsey_free_tickets(Whupsey *w, WhupseyTicket *tickets, int ntickets)
{
    int i, j;

    for (i = 0; i < ntickets; i++)
    {
        free(tickets[i].summary);
        free(tickets[i].requester);
        free(tickets[i].owner);
        free(tickets[i].comment);

        if (tickets[i].nattributes > 0)
        {
            for (j = 0; j < tickets[i].nattributes; j++)
                free(tickets[i].attributes[j].value);

            free(tickets[i].attributes);
        }
    }

    free(tickets);
}

void whupsey_clear_queries(Whupsey *w, WhupseyQuery *queries, int nqueries)
{
    int i;

    for (i = 0; i < nqueries; i++)
    {
        if (queries[i].get_details)
            whupsey_free_tickets(w, queries[i].tickets, queries[i].ntickets);
        else
            free(queries[i].ticket_ids);

        queries[i].ticket_ids = NULL;
        queries[i].tickets    = NULL;
        queries[i].ntickets   = 0;
    }
}

static void whupsey_free_schema(Whupsey *w)
{
    whupsey_free_types(w);
    whupsey_free_queues(w);
}

static void whupsey_free_types(Whupsey *w)
{
    int i;

    if (w->types != NULL)
    {
        for (i = 0; i < w->ntypes; i++)
        {
            free(w->types[i].name);
            free(w->types[i].description);
            whupsey_free_states(w, w->types[i].states, w->types[i].nstates);
            whupsey_free_priorities(w, w->types[i].priorities, w->types[i].npriorities);
        }

        free(w->types);
    }

    w->ntypes = 0;
    w->types  = NULL;
}

static void whupsey_free_states(Whupsey *w, WhupseyState *states, int nstates)
{
    int i;

    if (!states)
        return;

    for (i = 0; i < nstates; i++)
    {
        free(states[i].name);
        free(states[i].description);
    }

    free(states);
}

static void whupsey_free_priorities(Whupsey *w, WhupseyPriority *priorities, int npriorities)
{
    int i;

    if (!priorities)
        return;

    for (i = 0; i < npriorities; i++)
    {
        free(priorities[i].name);
        free(priorities[i].description);
    }

    free(priorities);
}

static void whupsey_free_versions(Whupsey *w, WhupseyVersion *versions, int nversions)
{
    int i;

    if (!versions)
        return;

    for (i = 0; i < nversions; i++)
    {
        free(versions[i].name);
        free(versions[i].description);
    }

    free(versions);
}

static void whupsey_free_queues(Whupsey *w)
{
    int i;

    if (w->queues != NULL)
    {
        for (i = 0; i < w->nqueues; i++)
        {
            free(w->queues[i].name);
            free(w->queues[i].description);
            free(w->queues[i].types);

            whupsey_free_versions(w, w->queues[i].versions, w->queues[i].nversions);
        }

        free(w->queues);
    }

    w->nqueues = 0;
    w->queues  = NULL;
}

